# -*- coding: utf-8 -*-
"""
Encapsulates objective functions for TSP
"""

from abc import ABC, abstractmethod
import numpy as np

def symmetric_tour_list(n_cities, start_index = 0):
    """
    Returns a numpy array representiung a symmetric tour of cities
    length = n_cities + 1
    First and last cities are index 'start_index' e.g. for
    start_index = 0 then for 5 cities tour = 
    [0, 1, 2, 3, 4, 0]
    """
    tour = [x for x in range(n_cities)]
    tour.remove(start_index)
    tour.insert(0, start_index)
    tour.append(start_index)
    return tour


def tour_cost(tour, matrix):
    """
    The total distance in the tour.
    
    @tour numpy array for tour
    @matrix numpy array of costs in travelling between 2 points.
    """
    cost = 0
    for i in range(len(tour) - 1):
        cost += matrix[tour[i]][tour[i+1]]
        
    return cost


class AbstractObjective(ABC):   
    @abstractmethod
    def evaluate(self, solution):
        pass


class SimpleTSPObjective(AbstractObjective):
    '''
    Simple objective for the Symmetric TSP
    Evaluates the cost of a tour.
    '''
    def __init__(self, matrix):
        '''
        Constructor

        Parameters:
        -------
        matrix - numpy.array, matrix (2D array) representing the 
        edge costs between each city.
        '''
        self._matrix = matrix
    
    def evaluate(self, tour):
        """
        The eucidean total distance in the tour.
        
        Parameters: 
        --------
        tour -  numpy.array, vector (1D array) representing tour
                e.g. [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        Returns:
        -------
        float - cost of tour.  This is the euclidean difference between each
        city in @tour with the addition of looping back from the final city to the 
        first.

        """
        cost = 0.0
        for i in range(len(tour) - 1):
            cost += self._matrix[tour[i]][tour[i+1]]

        cost += self._matrix[tour[len(tour)-1]][tour[0]]
            
        return cost
    

class OptimisedSimpleTSPObjective(AbstractObjective):
    '''
    Simple objective for the Symmetric TSP
    Evaluates the cost of a tour.
    '''
    def __init__(self, matrix):
        '''
        Constructor

        Parameters:
        -------
        matrix - numpy.array, matrix (2D array) representing the 
        edge costs between each city.
        '''
        self._matrix = matrix
    
    def evaluate(self, tour):
        """
        The eucidean total distance in the tour.
        
        Parameters: 
        --------
        tour -  numpy.array, vector (1D array) representing tour
                e.g. [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        Returns:
        -------
        float - cost of tour.  This is the euclidean difference between each
        city in @tour with the addition of looping back from the final city to the 
        first.

        """
        #create indexer for lookup in matrix
        indexer = np.concatenate([tour[:-1, None], tour[1:, None]], axis=1)
        return_trip = np.array([tour[-1], tour[0]]).reshape(1, -1)
        indexer = np.concatenate([indexer, return_trip])

        #fancy indexing to find costs and summation
        return self._matrix[indexer[:,0], indexer[:,1]].sum()


class CVRPObjective(AbstractObjective):
    '''
    Objective for capacitated vehicle routing
    problem.  Assumes vehicles are all of same type
    '''
    def __init__(self, matrix, warehouse, demand, capacity):
        '''
        Constructor

        Parameters:
        -------
        matrix - numpy.array, matrix (2D array) 
            representing the edge costs between each city.

        warehouse - int
            warehouse identifier or index

        demand - dict
            demand at each city

        capacity - float
            maximum capacity of each vehicle 
        '''
        self._matrix = matrix
        self._warehouse = warehouse
        self._demand = demand
        self._capacity = capacity 
    
    def evaluate(self, tour):
        """
        The eucidean total distance in the tour.
        
        Parameters: 
        --------
        tour -  numpy.array, vector (1D array) representing vehicle routes
                e.g. [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        Returns:
        -------
        float - cost of tour.  This is the euclidean difference between each
        city in @tour with the addition of looping back from the final city to the 
        first.
        """
        cost = 0.0
        routes = self._convert_tour_to_routes(tour)
        
        for subtour in routes:
            cost += self._subroute_cost(subtour)
        return cost

    def _subroute_cost(self, tour):
        '''
        Dev note: Would be more efficient to calculate
        route cost at same time as constructing them
        but leave for now...
        '''
        cost = 0.0
        for i in range(len(tour) - 1):
            cost += self._matrix[tour[i]][tour[i+1]]

        cost += self._matrix[tour[len(tour)-1]][tour[0]]    
        return cost
        
    def _convert_tour_to_routes(self, tour):
        routes = []
        subroute = [self._warehouse]
        total_load = 0.0
        for city in tour:
            if total_load + self._demand[city] <= self._capacity:
                subroute.append(city)
                total_load += self._demand[city]
            else:
                routes.append(subroute)
                subroute = [self._warehouse, city]
                total_load = self._demand[city]
        return routes


class CVRPUnitDemandObjective(AbstractObjective):
    '''
    Objective for capacitated vehicle routing
    problem.  Assumes vehicles are all of same type
    '''
    def __init__(self, matrix, warehouse, capacity, enforce_symmetric=True):
        '''
        Constructor

        Parameters:
        -------
        matrix - numpy.array, matrix (2D array) 
            representing the edge costs between each city.

        warehouse - int
            warehouse identifier or index

        capacity - float
            maximum capacity of each vehicle 

        enforce_symmetric optional (default = True)
            road networks are rarely symmetrical.  If simplifying
            to symmetric problem then set this to true and it 
            will use the top half of the travel matrix.
        '''
        self._matrix = np.asarray(matrix)
        self._warehouse = warehouse
        self._capacity = capacity 
        self._symmetric = enforce_symmetric
    
    def evaluate(self, tour):
        """
        The eucidean total distance in the tour.
        
        Parameters: 
        --------
        tour -  numpy.array, vector (1D array) representing vehicle routes
                e.g. [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        Returns:
        -------
        float - cost of tour.  This is the euclidean difference between each
        city in @tour with the addition of looping back from the final city to the 
        first.
        """
        cost = 0.0
        routes = self._convert_tour_to_routes(tour[1:])
        
        for subtour in routes:
            cost += self._subroute_cost(self._warehouse, subtour)
        return cost

    def _subroute_cost(self, warehouse_index, tour):
        '''
        Dev note: Would be more efficient to calculate
        route cost at same time as constructing them
        but leave for now...
        '''
        cost = 0.0
        #add in warehouse at start of tour
        tour = np.concatenate([np.asarray([warehouse_index]), 
                               tour, 
                               np.asarray([warehouse_index])])

        #loop through and calculate costs
        for i in range(len(tour) - 1):
            city_1, city_2 = tour[i], tour[i+1]
            if self._symmetric:
                city_1, city_2 = self._enforce_symmetry(city_1, city_2)
            #cities the other way around from pandas as numpy matrix
            cost += self._matrix[city_2][city_1]
            #cost += self._matrix2.iat[city_1, city_2]
            
        #return to warehouse
        #cost += self._matrix[tour[len(tour)-1]][tour[0]]    
        return cost
    
    def _enforce_symmetry(self, city_1, city_2):
        '''
        Road travel distances/times are not symmetric 
        This makes sure that the index row city is always the lowest
        index in the lookup.  Works for both postcode sector strings
        and indexes

        Parameters:
        -----
        city_1 - int  or str
            index/postcode sector of the first city
        
        city_2 - int 
            index/postcode sector of the second city

        Returns:
        -------
        Tuple 
        '''
        if city_1 > city_2:
            city_1, city_2 = city_2, city_1
        return city_1, city_2


    def _convert_tour_to_routes(self, tour):
        '''
        Easy as demand comes in single units and
        capacity is an integer
        '''
        n_cities = len(tour)
        capacity = int(self._capacity)
        splits = [i for i in range(capacity, n_cities, capacity)]
        routes = np.split(tour, splits)
        return routes


# -*- coding: utf-8 -*-
"""
hill-climbing local search and associated classes/functions

@author: Tom Monks
"""

import numpy as np
import time

class HillClimberRandomRestarts:
    '''
    Implementation of Hill-Climbing with Random Restarts
    '''
    def __init__(self, objective, localsearch, init_solution,
                 maxiter=20, random_seed=None):
        self._objective = objective
        
        self.set_init_solution(init_solution)
        self._maxiter = maxiter
        self._rng = np.random.default_rng(random_seed)
        self.localsearch = localsearch
        
    def set_init_solution(self, solution):  
        self.solution = solution
        self.best_solutions = [solution]
        
    def solve(self):
        
        local_solutions = []
        local_costs = []
                
        for i in range(self._maxiter):
            
            #create the hill climber
            self.localsearch.set_init_solution(np.copy(self.solution))
        
            #complete local search and get new solution
            self.localsearch.solve()
            
            #store local solution
            local_solutions.append(np.copy(self.localsearch.best_solutions[0]))
            local_costs.append(self.localsearch.best_cost)
                            
            #random restart
            self._rng.shuffle(self.solution)
            
        # best solution
        best_index = np.argmax(np.array(local_costs))
        self.best_cost = local_costs[best_index]
        self.best_solutions = [local_solutions[best_index]]

        

class HillClimber:
    '''
    Simple first improvement hill climbing algorithm
    
    '''
    def __init__(self, objective, init_solution, tweaker, maximisation=True,
                 time_limit=None):
        '''
        Constructor
        
        Params:
        ------
        objective:object
            optimisation target
        
        init_solution: np.ndarray
            numpy representation of solution
            
        tweaker: object
            tweak operation for hill climber
            
        maximisation: bool, optional (default=True)
            Is this a max or min optimisation?
            
        time_limit: float, optional (default=None)
            If set to float hill climbing termates when time limit is reached.
        '''
        self._objective = objective
        
        if maximisation:
            self._negate = 1.0
        else:
            self._negate = -1.0
        
        self.set_init_solution(init_solution)
        self._tweaker = tweaker
        
        #hill climbing time limit
        if time_limit is None:
            self._time_limit = np.Inf
        else:
            self._time_limit = time_limit
 
    def set_init_solution(self, solution):  
        self.solution = solution
        self.best_solutions = [solution]
        self.best_cost = self._objective.evaluate(solution) * self._negate 
               
    def solve(self):
        '''
        Run first improvement hill climbing
        
        Returns:
        --------
        None
        '''
        improvement = True
        start = time.time()
        
        while improvement and ((time.time() - start) < self._time_limit):
            improvement = False
            
            for city1 in range(1, len(self.solution) - 1):
                #print("city1: {0}".format(city1))
                for city2 in range(city1 + 1, len(self.solution) - 1):
                    #print("city2: {0}".format(city2))
                    
                    self._tweaker.tweak(tour=self.solution, start_index=city1, 
                                        end_index=city2)

                    neighbour_cost = \
                        self._objective.evaluate(self.solution) * self._negate

                    if (neighbour_cost > self.best_cost):
                        self.best_cost = neighbour_cost
                        self.best_solutions = [self.solution]
                        improvement = True
                    else:
                        self._tweaker.tweak(self.solution, city1, city2)
        



class TweakTwoOpt(object):
    '''
    Perform a 2-Opt swap for a tour (reverse a section of the route)
    '''
    def __init__(self, random_seed=None):
        '''
        Constructor method
        
        Params:
        ------
        random_seed: int, optional (default=None)
            control sampling in the algorithm.  By default is set
            to None i.e. random init.
        '''
        self.rng = np.random.default_rng(random_seed)
    
    def tweak(self, tour, start_index, end_index):
        '''
        Perform a 2-Opt swap for a tour (reverse a section of the route)
        
        Params:
        ------
        tour - np.ndarray
            vector representing tour
            
        start_index: int
            Index of first city in section of tour to reverse
            
        end_index: int
            Index of second city in section of tour to reverse
            
        
        '''
        self.reverse_section(tour, start_index, end_index)

    def reverse_section(self, tour, start, end):
        """
        Reverse a slice of the @tour elements between
        @start and @end. Note Operation happens in place.

        Params:
        --------
        tour - np.array, 
            vector representing a solution
            
        start - int, 
            start index of sublist (inclusive)
            
        end - int, 
            end index of sublist (inclusive)

        """
        tour[start:end] = tour[start:end][::-1]
        
        

class SimpleTweak(object):
    '''
    Perform a simple swap of two cities in a tour
    '''
    def __init__(self, random_seed=None):
        '''
        Constructor method
        
        Params:
        ------
        random_seed: int, optional (default=None)
            control sampling in the algorithm.  By default is set
            to None i.e. random init.
        '''
        self.rng = np.random.default_rng(random_seed)
    
    def tweak(self, tour, start_index, end_index):
        '''
        Swap two cities in tour
        
        Params:
        ------
        tour - np.ndarray
            vector representing tour
            
        start_index: int
            Index of first city to swap
            
        end_index: int
            Index of second city to swap
        
        '''
        tour[start_index], tour[end_index] = tour[end_index], tour[start_index]
        
        
        


class RandomTweakTwoOpt(object):
    '''
    Perform a 2-Opt swap for a tour (reverse a section of the route)
    Randomly selects section
    '''
    def __init__(self, random_seed=None):
        '''
        Constructor method
        
        Params:
        ------
        random_seed: int, optional (default=None)
            control sampling in the algorithm.  By default is set
            to None i.e. random init.
        '''
        self.rng = np.random.default_rng(random_seed)
        self._tweaker = TweakTwoOpt()
    
    def tweak(self, tour):
        '''
        Perform a 2-Opt swap for a tour (reverse a section of the route)
        
        Params:
        ------
        tour - np.ndarray
            vector representing tour
            
        Returns:
        --------
        np.ndarray
            tweaked tour
        
        '''
        sample = np.sort(self.rng.integers(0, high=len(tour), size=2))
        self.reverse_section(tour, tour[sample[0]], tour[sample[1]])
        return tour

    def reverse_section(self, tour, start, end):
        """
        Reverse a slice of the @tour elements between
        @start and @end. Note Operation happens in place.

        Params:
        --------
        tour - np.array, 
            vector representing a solution
            
        start - int, 
            start index of sublist (inclusive)
            
        end - int, 
            end index of sublist (inclusive)

        """
        tour[start:end] = tour[start:end][::-1]







    





# -*- coding: utf-8 -*-
"""
Provides simple functions for reading in TSP problems from known test problems.
Text file format from TSPLib: https://www.iwr.uni-heidelberg.de/groups/comopt/software/TSPLIB95/
Requires knowledge about the volumne of meta_data.

Slight issue - numpy.loadtxt does not like EOF. This is what is in the tsp files
my workaround is to delete EOF from the file!!
"""

import numpy as np

def read_coordinates(file_path, md_rows):
    """
    Return numpy array of city coordinates. 
    Excludes meta data (first 6 rows) and first column (city number)
    """
    return np.loadtxt(file_path, skiprows = md_rows, usecols = (1,2))

def read_meta_data(file_path, md_rows):
    """
    Returns meta dat/paraetmers a from specified tsp file
    """
    with open(file_path) as tspfile:
        return [next(tspfile).rstrip('\n') for x in range(md_rows - 1)]
        
        


    
    




# -*- coding: utf-8 -*-
"""
Created on Mon Sep 18 07:48:50 2017

@author: tm3y13
"""
from metapy.tsp.objective import tour_cost



class NearestNeighbour(object):
    
    def __init__(self, cities, matrix):
        self.cities = cities
        self.matrix = matrix
        self.solution = []
        self.best_cost = 99999
        self.best_solution = self.solution
        
    def solve(self):
        """
        Constructs a tour based on nearest neighbour method.
        Assume that first city in tour is base city.
        """
        
        from_city = self.cities[0]
        
        self.solution.append(from_city)
                
        for i in range(len(self.cities) - 2):
            to_city = self.closest_city_not_in_tour(from_city)
            self.solution.append(self.cities[to_city])
            #print(self.solution)
            from_city = to_city
        
        self.solution.append(self.cities[0])
        self.best_cost = tour_cost(self.solution, self.matrix)
        self.best_solution = self.solution
                   
            
                    
    def closest_city_not_in_tour(self, from_city):
        min_cost = 999
        min_index = from_city
        
        for to_city in range(len(self.cities) - 1):
            
            if (min_cost > self.matrix[from_city][to_city]):
                              
                if (self.cities[to_city] not in self.solution):
                    min_index = to_city
                    min_cost = self.matrix[from_city][to_city]
        
        return min_index
        
        
class FurthestInsertion(object):
    """
    Furthest insertion contruction heuristic.  
    Finds city furthest away from any point on the subtour.
    Finds a place to insert into tour that minimises the increase in length
    
    Notes: code needs a tidy - TM.
    """
    def __init__(self, cities, matrix):
        self.cities = cities
        self.matrix = matrix
        self.solution = []
        self.best_cost = 0
        self.best_solution = self.solution
        
    
    def solve(self):
        from_city = self.cities[0]
        
        self.solution.append(from_city)
        print(self.solution)
        
        #step 2
        to_insert = self.select_city(self.solution)
        result = self.furthest_city_index(from_city)
        self.solution.append(result[1])
        self.best_cost = result[0]
        
        print(self.solution)
        
        
        #loop until full tour
        for i in range(len(self.cities) - 3):
            
            #find city to insert
            to_insert = self.select_city(self.solution)
            
            #find insertion point
            self.insert_link(to_insert)
        
        self.solution.append(self.solution[0])
        print(self.solution)
        
        self.best_cost = tour_cost(self.solution, self.matrix)
        
    def select_city(self, sub_solution):           
        max_cost = 0
        index = 0
        #[self.farthest_city_index(from_city) for from_city in sub_solution]
        
        for from_city in sub_solution:
            result =  self.furthest_city_index(from_city)
            
            if max_cost < result[0]:
                max_cost = result[0]
                index = result[1]
                
        return index
               
               
        
    def furthest_city_index(self, from_city):
        cost = 0
        index = 0
        for to_city in self.cities[1:len(self.cities)]:
            #print("furthest city from {0}".format(from_city))
            if(self.matrix[from_city][to_city] > cost):
                
                if(to_city not in self.solution):
                    
                    cost = self.matrix[from_city][to_city]
                    index = to_city
        
        return [cost, index]
                
        
    def insert_link(self, to_insert):
        """
        Insert edge in between cities to achieve
        minimum increase in length
        """
        
        #print("to insert {0}".format(to_insert))
        deltas = [[self.calculate_delta(pos, to_insert), pos] for pos in range(1, len(self.solution))]
        min_insertion = min(deltas)
        self.solution.insert(min_insertion[1], to_insert)   
        print(self.solution)
        self.best_cost += min_insertion[0]
        
        
    def calculate_delta(self, i, to_insert):
        """
        Calculate the difference between current tour length 
        and additional after potential insert of @to_insert
        @to_insert - the city to insert 
        """
        #print("C_ik = {0}".format(self.matrix[i][to_insert]))
        #print("C_ki = {0}".format(self.matrix[to_insert][i+1]))
        #print("C_ij = {0}".format(self.matrix[i][i+1]))
        
        return self.matrix[i][to_insert] + self.matrix[to_insert][i+1] \
            - self.matrix[i][i+1]                     
            
        




# -*- coding: utf-8 -*-
"""
Created on Thu Sep 14 15:57:46 2017

@author: tm3y13
"""

from random import shuffle
import numpy as np
from metapy.evolutionary.evolutionary import AbstractPopulationGenerator

def random_tour(tour):
    """
    Initial solution to tour is psuedo random
    """
    rnd_tour = tour[1:len(tour)-1]
    base_city = tour[0]
    shuffle(rnd_tour)
    rnd_tour.append(base_city)
    rnd_tour.insert(0, base_city)
    return rnd_tour


class TSPPopulationGenerator(AbstractPopulationGenerator):
    def __init__(self, cities_in_tour):
        self._cities_in_tour = cities_in_tour

    def generate(self, population_size):
        '''
        Generate a list of @population_size tours.  Tours
        are randomly generated and unique to maximise
        diversity of the population.

        Parameters:
        ---------
        population_size -- the size of the population

        Returns:
        ---------
        np.array. matrix size = (population_size, len(tour)). Contains
                the initial generation of tours
        '''

        population = {}

        #return data as
        population_arr = np.full((population_size, len(self._cities_in_tour)), 
            -1, dtype=np.byte)

        i = 0
        while i < population_size:
            #sample a permutation
            new_tour = np.random.permutation(self._cities_in_tour)
            
            #check its unique to maximise diversity
            if str(new_tour) not in population:
                population[str(new_tour)] = new_tour
                i = i + 1

        #save unique permutation
        population_arr[:,] = list(population.values())

        return population_arr


def initiation_population(population_size, tour):
    '''
    Generate a list of @population_size tours.  Tours
    are randomly generated and unique to maximise
    diversity of the population.

    Parameters:
    ---------
    population_size -- the size of the population

    Returns:
    ---------
    np.array. matrix size = (population_size, len(tour)). Contains
              the initial generation of tours
    '''

    population = {}
    #for i in range(population_size):
    i = 0
    while i < population_size:
        #some code is legacy and uses python
        #lists instead of numpy arrays... to fix!
        #the random tour bit varies between problem...
        #new_tour = random_tour(tour)
        new_tour = tour.copy()
        np.random.shuffle(new_tour)
        
        #return data as
        population_arr = np.full((population_size, len(tour)), -1, dtype=np.byte)

        if str(new_tour) not in population:
            population[str(new_tour)] = np.array(new_tour)
            i = i + 1

    population_arr[:,] = list(population.values())

    return population_arr
    


# -*- coding: utf-8 -*-
"""
Useful but misc utility functions for solving TSP problems.
"""

def trim_base(full_tour):
    """
    Trims start/end city from tour and returns the trimmed
    tour and the base city.
    
    E.g. Passed a full symmetric tour [0, 1, 2, 0] 
    trim_base returns two objects:
        trimmed_tour = [1, 2]
        base_city 0
    """
    base_city = full_tour[0]
    trimmed_tour = full_tour[1:len(full_tour)-1]
    return trimmed_tour, base_city


def append_base(trimmed_tour, base_city):
    """
    Appends a common start/end to a tour
    E.g. when passed
    trimmed_tour = [1,2]
    base_city = 0
    The func returns [0, 1, 2, 0]
    """
    trimmed_tour.append(base_city)
    trimmed_tour.insert(0, base_city)
    return trimmed_tour




# -*- coding: utf-8 -*-
"""
Created on Sat Sep 16 20:59:46 2017

@author: tm3y13
"""

from abc import ABC, abstractmethod
import numpy as np

from metapy.tsp.init_solutions import random_tour


class ILSPertubation(ABC):
    @abstractmethod
    def perturb(self, tour):
        pass


class ILSHomeBaseAcceptanceLogic(ABC):
    @abstractmethod
    def new_home_base(self, home_base, home_cost, candidate, candidate_cost):
        pass
    
    
class AbstractCoolingSchedule(ABC):
    '''
    Encapsulates a cooling schedule for
    a SA algorithm.
    Abstract base class.
    Concrete implementations can be
    customised to extend the range of
    cooling schedules available to the SA.
    '''
    def __init__(self, starting_temp):
        self.starting_temp = starting_temp

    @abstractmethod
    def cool_temperature(self, k):
        pass
    

class HigherQualityHomeBase(ILSHomeBaseAcceptanceLogic):
    '''
    Accept if candidate is better than home_base
    '''

    def new_home_base(self, home_base, home_cost, candidate, candidate_cost):
        if candidate_cost > home_cost:
            return candidate, candidate_cost
        else:
            return home_base, home_cost

class RandomHomeBase(ILSHomeBaseAcceptanceLogic):
    '''
    Random walk homebase
    '''
    def new_home_base(self, home_base, home_cost, candidate, candidate_cost):
        return candidate, candidate_cost
    
    
class EpsilonGreedyHomeBase(ILSHomeBaseAcceptanceLogic):
    '''
    Accept if candidate is better than home_base of 
    if sampled u > epsilon otherwise explore.
    '''
    def __init__(self, epsilon=0.2, exploit=None, explore=None):
        self.epsilon = epsilon
        if exploit is None:
            self.exploit = HigherQualityHomeBase()
        else:
            self.exploit = exploit

        if explore is None:
            self.explore = RandomHomeBase()
        else:
            self.explore = explore

    def new_home_base(self, home_base, home_cost, candidate, candidate_cost):
        
        u = np.random.rand()

        if u > self.epsilon:
            return self.exploit.new_home_base(home_base, 
                                              home_cost, 
                                              candidate, 
                                              candidate_cost)
        else:
            return self.explore.new_home_base(home_base, 
                                              home_cost, 
                                              candidate, 
                                              candidate_cost)
            
            
class AnnealingEpsilonGreedyHomeBase(ILSHomeBaseAcceptanceLogic):
    '''
    Accept if candidate is better than home_base of 
    if sampled u > epsilon otherwise explore.
    
    Epsilon decreases from 1 to 0.
    
    '''
    def __init__(self, maxiter_per_temp=1, annealing=None, exploit=None, 
                 explore=None, verbose=False):
        
        #create epsilon greedy hombase
        self.accept = EpsilonGreedyHomeBase(1.0, exploit, explore)
        self.actions = 0
        self.verbose = verbose
        self.maxiter_per_temp = maxiter_per_temp
        
        #scheme to reduce epsilon
        if annealing is None:
            #default is basic approach from SA. 0.95 * temp
            self.annealing = ExponentialCoolingSchedule(1.0)
            
        else:
            self.annealing = annealing
        
        #iterations at current temperature
        self.iters_at_temp = 0
       

    def new_home_base(self, home_base, home_cost, candidate, candidate_cost):
        
        #new home base
        candidate, candidate_cost = \
            self.accept.new_home_base(home_base, 
                                      home_cost, 
                                      candidate, 
                                      candidate_cost)
       
        #track number of updates to home base and iters at temp
        self.iters_at_temp += 1
        
        
        if self.iters_at_temp == self.maxiter_per_temp: 
            #update epsilon
            self.actions += 1
            self.accept.epsilon = self.annealing.cool_temperature(self.actions) 
            self.iters_at_temp = 0
            if self.verbose:
                print(f'epsilon: {self.accept.epsilon}')
        
        return candidate, candidate_cost




class ExponentialCoolingSchedule(AbstractCoolingSchedule):
    '''
    Expenontial Cooling Scheme.
    Source:
    https://uk.mathworks.com/help/gads/how-simulated-annealing-works.html
    '''
    def cool_temperature(self, k):
        '''
        Cool the temperature using the scheme
        T = T0 * 0.95^k.
        Where
        T = temperature after cooling
        T0 = starting temperature
        k = iteration number (within temperature?)
        Keyword arguments:
        ------
        k -- int, iteration number (within temp?)
        Returns:
        ------
        float, new temperature
        '''
        return self.starting_temp * (0.95**k)
    
    
class TempFastCoolingSchedule(AbstractCoolingSchedule):
    '''
    Cooling the temperature quickly!
    Source:
    https://uk.mathworks.com/help/gads/how-simulated-annealing-works.html
    '''
    def cool_temperature(self, k):
        '''
        Cool the temperature using the scheme
        T = T0 / k.
        Where
        T = temperature after cooling
        T0 = starting temperature
        k = iteration number 
        Keyword arguments:
        ------
        k -- int, iteration number (within temp?)
        Returns:
        ------
        float, new temperature
        '''
        return self.starting_temp / k
    
    
    
class BoltzCoolingSchedule(AbstractCoolingSchedule):
    '''
    Boltzman Cooling Scheme.
    Source:
    https://uk.mathworks.com/help/gads/how-simulated-annealing-works.html
    '''
    def cool_temperature(self, k):
        '''
        Cool the temperature using the scheme
        T = T0 / ln(k).
        Where
        T = temperature after cooling
        T0 = starting temperature
        k = iteration number (within temperature?)
        Keyword arguments:
        ------
        k -- int, iteration number
        Returns:
        ------
        float, new temperature
        '''
        return self.starting_temp / np.log(k)
    
    

class DoubleBridgePertubation(ILSPertubation):
    '''
        Perform a random 4-opt ("double bridge") move on a tour.
        
         E.g.
        
            A--B             A  B
           /    \           /|  |\
          H      C         H------C
          |      |   -->     |  |
          G      D         G------D
           \    /           \|  |/
            F--E             F  E
        
        Where edges AB, CD, EF and GH are chosen randomly.

    '''

    def perturb(self, tour):
        '''
        Perform a random 4-opt ("double bridge") move on a tour.
        
        Returns:
        --------
        numpy.array, vector. representing the tour

        Parameters:
        --------
        tour - numpy.array, vector representing tour between cities e.g.
               [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        '''

        n = len(tour)
        end = int(n/4)+1
        end = int(n/3)+1
        pos1 = np.random.randint(1, end) 
        pos2 = pos1 + np.random.randint(1, end) 
        pos3 = pos2 + np.random.randint(1, end) 

        #print(tour[:pos1], tour[pos1:pos2], tour[pos2:pos3], tour[pos3:])

        p1 = np.concatenate((tour[:pos1] , tour[pos3:]), axis=None)
        p2 = np.concatenate((tour[pos2:pos3] , tour[pos1:pos2]), axis=None)
        #this bit will need updating if switching to alternative operation.
        #return np.concatenate((p1, p2, p1[0]), axis=None)
        return np.concatenate((p1, p2), axis=None)
    
    
class TabuDoubleBridgeTweak(object):
    '''
    Wraps a DoubleBridgePertubation with a limited size memory
    
    The memory acts a tabu list and prevents the tweak from returning
    to previous solutions
    
    '''
    def __init__(self, tabu_size, init_solution):
        '''
        constructor
        
        Params:
        -------
        tabu_size: int
            Size of tabu list
            
        init_solution: np.ndarray
            First solution to add to the memory
            
        '''
        self.tabu_size = tabu_size
        
        #double bridge tweak
        self.perturber = DoubleBridgePertubation()
        
        self.history = [init_solution.tolist()]
       
        
    def perturb(self, tour):
        '''
        apply the perturbation to the tour.
        
        Params:
        ------
        tour: np.ndarray
            vector representing tour
            
        Returns:
        -------
        np.ndarray
            perturbed tour.
        '''
        candidate = self.perturber.perturb(tour)
    
        while candidate.tolist() in self.history:
            print('loop')
            candidate = self.perturber.perturb(tour)
            
        self.history.append(candidate.tolist())
        if len(self.history) > self.tabu_size:
            del self.history[0]
        
        return candidate


if __name__ == '__main__':
    tour = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    d = DoubleBridgePertubation()
    d.perturb(tour)


class IteratedLocalSearch(object):
    '''
    Iterated Local Search metaheuristic 
    
    Intelligent search of local optima provided by a local search
    algorithm.
    
    Implemented to work with a fixed number of iterations.
    '''

    def __init__(self, local_search, accept=None, perturb=None, 
                 verbose=False):
        '''
        ILS constructor Method

        Parameters:
        --------
        
        local_search: object,
            hill climbing solver or similar
            
        accept: ILSHomeBaseAcceptanceLogic, optional (default=RandomHomeBase())
            contains the logic for accepting a new homebase
        
        perturb: ILSPertubation, 
            contains the logic for pertubation from the local optimimum in each 
            iteration
            
        verbose: bool, optional (default=False)
            When verbose=True the results of each iteration are printed
            Useful for learning how the algorithm works.
        '''

        self._local_search = local_search
        if accept == None:
            self._accepter = RandomHomeBase()
        else:
            self._accepter = accept

        if perturb == None:
            self._perturber = DoubleBridgePertubation()
        else:
            self._perturber = perturb
            
        self.best_solutions = []
        self.best_cost = self._local_search.best_cost
        self.verbose = verbose
        
        if verbose:
            print(f'ITERATED LOCAL SEARCH')
            print(f'Initial solution cost: {self.best_cost}')
                
    def run(self, n):
        """
        Re-run solver n times using a different initial solution
        each time.  Init solution is generated randomly each time.

        The potential power of iterated local search lies in its biased 
        sampling of the set of local optima.

        """

        current = self._local_search.solution
                
        home_base = current
        home_base_cost = self._local_search.best_cost
        self.best_cost = home_base_cost 
        
        self.best_solutions.append(current)
                
        for x in range(n):

            #Hill climb from new starting point
            self._local_search.set_init_solution(current)
            self._local_search.solve()
            current = self._local_search.best_solutions[0]

            #best cost
            iteration_best_cost = self._local_search.best_cost
            
            if iteration_best_cost > self.best_cost:
                self.best_cost = iteration_best_cost
                self.best_solutions = self._local_search.best_solutions

            elif iteration_best_cost == self.best_cost:
                [self.best_solutions.append(i) 
                     for i in self._local_search.best_solutions]

            #update homebase
            home_base, home_base_cost = \
                self._accepter.new_home_base(home_base, home_base_cost, 
                                             current, iteration_best_cost)
            if self.verbose:
                print(f'{x+1}', end=': ')
                print(f'iteration cost: {iteration_best_cost}', end='; ')
                print(f'homebase cost: {home_base_cost}', end='; ')
                print(f'best found {self.best_cost}')
                
            current = self._perturber.perturb(home_base)
        
    def get_best_solutions(self):
        return self.best_cost, self.best_solutions

  
            
            
            


'''
Greedy Randomised Adaptive Search Procedure

classes and functions.
'''

import numpy as np
import time

class FixedRCLSizer:
    '''
    Fixed sized RCL list.
    
    When r = 1 then greedy
    When r = len(tour) then random
    '''
    def __init__(self, r):
        self.r = r
        
    def get_size(self):
        '''
        Returns an int representing the size of the required RCL
        '''
        return self.r
    
class RandomRCLSizer:
    '''
    Probabilitic selection of the RCL size
    Uniform probability.
    '''
    def __init__(self, r_list, random_seed=None):
        self.r_list = r_list
        self.rng = np.random.default_rng(random_seed)
        
    def get_size(self, size=None):
        '''
        Returns a randomly selected RCL size
        '''
        return self.rng.choice(self.r_list, size=size)
    
class SemiGreedyConstructor:
    '''
    Semi-greedy construction of a tour.
    
    For a city i creates a restricted candidate list of size r
    i.e the r shortest distances from city i.  
    Next city is chosen with equal probability.
    Repeats until tour is constructed.
    '''
    def __init__(self, rcl_sizer, tour, matrix,
                 random_seed=None):
        '''
        Constructor
        
        Params:
        ------
        rcl_sizer: object
            sizes the restricted candidate list
        
        tour: np.ndarray
            vector of city indexes included in problem
            
        matrix: np.ndarray
            matrix of travel costs
            
        random_seed: int
            used to control sampling and provides a
            reproducible result.
        '''
        
        # size of rcl
        self.rcl_sizer = rcl_sizer
        
        # cities in a tour
        self.tour = tour
        
        # travel cost matrix
        self.matrix = matrix
        
        # create random number generator
        self.rng = np.random.default_rng(random_seed)
    
    def build(self):
        '''
        Semi-greedy contruction of tour
        
        Returns:
        --------
        np.array
        '''
        # first city in tour
        solution = np.array([self.tour[0]])    
        
        # it is an iterative (construction) procedure
        for i in range(len(self.tour)-1):
            # get the RCL size
            r = self.rcl_sizer.get_size()
            
            # get the RCL
            rcl = self.get_rcl(r, solution, solution[-1])
            
            # select the next city 
            next_city = self.random_from_rcl(rcl)
            
            # update the solution
            solution = np.append(solution, np.array([next_city]))
            
        return solution
    
    def get_rcl(self, r, solution, from_city):
        '''
        Restricted candidate list for final city in current solution
        
        Params:
        -------
        solution: np.ndarray
            vector of current partially constructed solution
            
        from_city: int
            index of city used to construct rcl.
        
        Returns:
        -------
        np.array
        '''
        # get indexes of cities not in solution
        mask = self.tour[~np.in1d(self.tour, solution)]
        
        # get indexes of r smallest travels costs 
        if mask.shape[0] > r:
            # partition the vector for remaining cities - faster than sorting 
            idx = np.argpartition(self.matrix[from_city][mask], 
                                  len(mask) - r)[-r:]
            rcl = mask[idx]
        else:
            # handle when r < n cities remaining 
            rcl = mask
        return rcl
    
    def random_from_rcl(self, rcl):
        '''
        Select a city at random from rcl.
        Return city index in self.matrix
        
        Params:
        -------
        rcl: np.ndarray
            restricted candidate list
            vector of candidate city indexes.
        '''
        return self.rng.choice(rcl)
    
    
class GRASP:
    '''
    Greedy Randomised Adaptive Search Procedure algorithm
    for the Travelling Salesman Problem.
    
    
    The class has the following properties
    .best: float
        the best cost
        
    .best_solution: np.ndarray
        the best tour found
    
    '''
    def __init__(self, constructor, local_search, max_iter=1000,
                 time_limit=np.inf):
        '''
        Constructor
        
        Parameters:
        ---------
        constructor: object
            semi-greedy construction heuristic
            
        local_search: object
            local search heuristic e.g. `HillClimber`
            
        max_iter: int, optional (default=1000)
            The maximum number of iterations (restarts) of GRASP
            
        time_limit: float64, optional (default=np.inf)
            The maximum allowabl run time.
            
        
        '''
        # semi greedy tour construction method
        self.constructor = constructor
        
        # local search procedure
        self.local_search = local_search
        
        # max runtime budget for GRASP
        self.max_iter = max_iter
        self.time_limit = time_limit
        
        # init solution 
        self.best_solution = None
        self.best = None
    
        
    def solve(self):
        '''
        Run GRASP
        
        Returns:
        -------
        None
        '''
        
        self.best_solution = None
        self.best = -np.inf
        
        i = 0
        start = time.time()
    
        while i < self.max_iter and ((time.time() - start) < self.time_limit):
            
            i += 1
            
            # construction phase
            solution = self.constructor.build()
            
            # Improve solution via local search
            self.local_search.set_init_solution(solution)
            self.local_search.solve()
            
            current_solution = self.local_search.best_solutions[0]
            current = self.local_search.best_cost
            
            # check if better than current solution
            if current > self.best:
                self.best = current
                self.best_solution = current_solution
    

    
class MonitoredLocalSearch:
    '''
    Extends a local search class and provides the observer pattern.
    An external object can observe the local search object and catch the
    termination event (end of local search).  The observer is notified and
    passed the results of the local search.

    Use cases: 
    ----------
    In GRASP this is useful for an algorithm sizing the RCL and learning 
    on average how different sizes of RCL perform.
    '''
    def __init__(self, local_search):
        '''
        Constructor:
        
        Params:
        ------
        local_search: Object
            Must implement .solve(), best_cost, best_solution
        '''
        self.local_search = local_search
        self.observers = []
    
    def register_observer(self, observer):
        '''
        register an object to observe the local search
        
        The observer should implement 
        local_search_terminated(*args, **kwargs)
        '''
        self.observers.append(observer)
        
    def set_init_solution(self, solution):
        '''
        Set the initial solution
        
        Params:
        --------
        solution: np.ndarray
            vector representing the initial solution
        '''
        self.local_search.set_init_solution(solution)
    
    def solve(self):
        '''
        Run the local search.  
        At the end of the run all observers are notified.
        '''    
        # run local search
        self.local_search.solve()
        
        # notify observers after search terminates.
        best = self.local_search.best_cost
        solution = self.local_search.best_solutions[0]
        self.notify_observers(best, solution)
        
    
    def notify_observers(self, *args, **kwargs):
        '''
        Observers must implement `local_search_terminated()`
        method.
        
        Params:
        ------
        *args: list
            variable number of arguments
            
        **kwargs: dict
            key word arguments
        '''
        for o in self.observers:
            o.local_search_terminated(*args, **kwargs)
    
    def _get_best_cost(self):
        '''
        best cost from internal local_search object
        '''
        return self.local_search.best_cost
    
    def _get_best_solutions(self):
        '''
        get best solutions from local_search object
        '''
        return self.local_search.best_solutions
    
    best_cost = property(_get_best_cost, doc='best cost')
    best_solutions = property(_get_best_solutions, doc='best solution')
    
    
class ReactiveRCLSizer:
    '''
    Dynamically update the probability of selecting a 
    value of r for the size of the RCL.
    
    Implements Reactive GRASP.
    
    '''
    def __init__(self, r_list, local_search, freq=None, random_seed=None):
        '''
        Constructor
        
        Params:
        -------
        r_list: list
            vector of sizes for RCL e.g. [1, 2, 3, 4, 5]
            
        local_search: MonitoredLocalSearch
            local_search to monitor
            
        freq: int, optional (default=None)
            Frequency in iterations at which the probabilities are updated.
            When set to None it defaults to the length of r_list * 2
            
        random_seed: int, optional (default=None)
            Control random sampling for reproducible result
        '''
        # list of r sizes
        self.r_list = r_list
        
        # set of indexes to work with probabilities
        self.elements = np.arange(len(r_list))
    
        # probability of choosing r (initially uniform)
        self.probs = np.full(len(r_list), 1/len(r_list))
        
        # mean performance of size r
        self.means = np.full(len(r_list), 1.0)
        
        # runs of size r
        self.allocations = np.full(len(r_list), 0)
        
        # local search to monitor
        self.local_search = local_search
        
        # frequency of updating probs
        if freq is None:
            self.freq = len(self.r_list)
        else:
            self.freq = freq
        
        # number of iterations within frequency
        self.iter = 0
        
        # current r index
        self.index = -1
        
        # to init run one of each r value
        self.init = True
        
        # imcumbent solution cost
        self.best_cost = -np.inf
        
        # register sizer as observer of the local search
        local_search.register_observer(self)
        
        # random no. gen
        self.rng = np.random.default_rng(random_seed)
    
    def local_search_terminated(self, *args, **kwargs):
        '''
        Termination of the local search
        '''
        # iteration complete
        self.iter += 1
        
        # get the best cost found in the iteration
        iter_cost = args[0]

        # record iteration took plaxe with index i
        self.allocations[self.index] += 1
        
        # update running mean
        mean_x = self.means[self.index]
        n = self.allocations[self.index]
        self.means[self.index] += (iter_cost - mean_x) / n
        
        self.update_r()
        
        # update incumbent cost if required
        if iter_cost > self.best_cost:
            self.best_cost = iter_cost
        
        # update probs if freq met.
        if self.iter >= self.freq and not self.init:
            self.iter = 0
            self.update_probability()
            
        
    def update_probability(self):
        '''
        Let $q_i = f^* / A_i$
        and $p_i = `\dfrac{q_i}{\sum_{j=1}^{m} q_j}$
        
        where
        
        $f^*$ is the incumbent (cost)
        $A_i$ is the mean cost found with r_i
        
        larger q_i indicates more suitable values of r_i
        '''
        q = self.best_cost / self.means
        self.probs = q / q.sum()
    
    def update_r(self):
        '''
        update the size of r
        
        Note that the implementation ensures that all r values are run
        for at least one iteration of the algorithm.
        '''
        # initial bit of logic makes sure there is at least one run of all probabilities
        if self.init:
            self.index += 1
            if self.index >= len(self.r_list):
                self.init = False
                self.index = self.rng.choice(self.elements, p=self.probs)
        else:
            self.index = self.rng.choice(self.elements, p=self.probs)
    
    def get_size(self):
        '''
        Return the selected size of the RCL
        
        The selection is done using a discrete distribution
        self.r_probs.
        '''
        return self.r_list[self.index]
    
    
class RandomPlusGreedyConstructor(SemiGreedyConstructor):
    '''
    Random + semi-greedy construction of a tour.
    
    The first n cities of a tour are randomly constructed.  
    The remaining cities are seleted using the standard semi-greedy approach.
    
    For a city i creates a restricted candidate list of size r
    i.e the r shortest distances from city i.  Next city is chosen
    with equal probability. 
    
    Repeats until tour is constructed.
    
    '''
    def __init__(self, rcl_sizer, tour, matrix, p_rand=0.2,
                 random_seed=None):
        '''
        RandomPlusGreedy Constructor method
        
        Params:
        ------
        rcl_sizer: object
            sizes the restricted candidate list
        
        tour: np.ndarray
            vector of city indexes included in problem
            
        matrix: np.ndarray
            matrix of travel costs
            
        p_rand: float, optional (default=0.2)
            Proportion of tour that is randomly constructed
            
        random_seed: int
            used to control sampling provides a
            reproducible result.
        '''
        
        # super class init
        super().__init__(rcl_sizer, tour, matrix,
                       random_seed)
        
        # proportion of tour that is randomly constructed
        self.p_rand = p_rand
        self.n_rand = int(p_rand * len(tour))
        self.n_greedy = len(tour) - self.n_rand - 1
        
    
    def build(self):
        '''
        Random followed by semi-greedy contruction of tour
        
        Returns:
        --------
        np.array
        '''
        # first city in tour
        solution = np.array([self.tour[0]])    
        # next n_rand cities are random
        rand = self.rng.choice(self.tour[1:], size=self.n_rand, replace=False)
        solution = np.append(solution, rand)
        
        # remaining cities are semi-greedy
        for i in range(self.n_greedy):
            r = self.rcl_sizer.get_size()
            rcl = self.get_rcl(r, solution, solution[-1])
            next_city = self.random_from_rcl(rcl)
            solution = np.append(solution, np.array([next_city]))
            
        return solution        

    
class ConstructorWithMemory:
    '''
    Provides a construction heuristic with a short term memory
    '''
    def __init__(self, constructor, memory_size=100):
        '''Constructor method
        
        Params:
        -------
        constructor: Object
            Implements build() and returns a solution
            
        memory_size, int, optional (default=100)
            size of tabu list
        '''
        self.constructor = constructor
        self.memory_size = memory_size
        # memory implemented as list
        self.history = []
        
    def build(self):
        '''
        Run the stochastic construction heuristic
        
        Re-runs heuristic if results is within memory
        
        Returns:
        --------
        np.ndarray
        '''
        solution = self.constructor.build()
        while str(solution) in self.history:
            solution = self.constructor.build()
        
        # if at capacity remove oldest solution
        if len(self.history) >= self.memory_size: 
            self.history.pop(0)
        
        self.history.append(str(solution))
        return solution
    
class EliteSet:
    '''
    Tracks and updates an elite set of solutions produced by a local search.
    '''
    def __init__(self, local_search=None, max_size=10, min_delta=0):
        '''
        Constructor
        Params:
        -------
        local_search: MonitoredLocalSearch
            The local search that produces candidates for the elite
            set.
        
        max_size: int, optional (default=10)
            maximum entries in the elite set

        min_delta: int, optional (Default=0)
            The min cardinality difference between tours to allow entry
            E.g. a = [1, 2, 3, 4, 5]; b = [1, 3, 4, 2, 5]. delta = 3.
            Vary delta > 0 to increase diversity (but may limit entry)
        '''
        if local_search is not None:
            self.local_search = local_search
            local_search.register_observer(self)
        
        self.min_delta = min_delta
        self.max_size = max_size
        
        # data structures for elite solutions
        self.solutions = None
        self.costs = None
        
        self.n_updates = 0
    
    @property
    def is_empty(self):
        return self.solutions is None
    
    def is_elite(self, solution):
        '''
        Is the solution a member of the elite set
        
        Params:
        ------
        solution: np.ndarray
            TSP solutution
            
        Returns:
        --------
        bool
        '''
        if self.solutions is None:
            return False
        else:
            result = np.where((self.solutions==solution).all(axis=1))[0]
            return len(result) > 0
    
    def local_search_terminated(self, *args, **kwargs):
        ''''
        Termination of the local search
        '''
        s = args[1]
        s_cost = args[0]
        self.update(s, s_cost)
            
    def init_elite_set(self, s, s_cost):
        '''
        Initalise the elite set
        '''
        self.solutions = np.array([s])
        self.costs = np.array([s_cost])
    
    def update(self, s, s_cost):
        '''
        Update the elite set to maximise performance and diversity

        Params:
        -------
        s: np.ndarray
            TSP tour

        s_cost: float
            TSP tour cost

        Returns:
        -------
        Tuple: np.ndarray, np.ndarray
            elite_set, elite_costs
        '''
        if self.solutions is None:
            self.init_elite_set(s, s_cost)
        
        elif len(self.solutions) < self.max_size:
            delta = (s != self.solutions).sum(axis=1).min()
            if delta > self.min_delta:
                self.solutions = np.append(self.solutions, [s], axis=0)
                self.costs = np.append(self.costs, [s_cost], axis=0)
        else:
            # aim to increase performance and diversity of elite set
            worst_cost = self.costs.min()
            delta = (s != self.solutions).sum(axis=1).min()

            if s_cost > worst_cost and delta > self.min_delta:
                replace_idx = self.costs.argmin()
                self.solutions[replace_idx] = s
                self.costs[replace_idx] = s_cost
                self.n_updates += 1
                
    def get_best_solution(self):
        '''Returns the best solution and cost
        from the elite set'''
        best_idx = np.array(self.costs).argmax()
        return self.solutions[best_idx], self.costs[best_idx]


import numpy as np
from abc import ABC, abstractmethod

class AbstractPopulationGenerator(ABC):
    @abstractmethod
    def generate(self, population_size):
        pass

class AbstractMutator(ABC):
    @abstractmethod
    def mutate(self, individual):
        pass

class AbstractEvolutionStrategy(ABC):
    @abstractmethod
    def evolve(self, population, fitness):
        pass

class AbstractSelector(ABC):
    @abstractmethod
    def select(self, population, fitness):
        pass

class AbstractCrossoverOperator(ABC):
    @abstractmethod
    def crossover(self, parent_a, parent_b):
        pass


class PartiallyMappedCrossover(AbstractCrossoverOperator):
    '''
    Partially Mapped Crossover operator
    '''
    def __init__(self):
        pass

    def crossover(self, parent_a, parent_b):
    
        child_a = self._pmx(parent_a.copy(), parent_b)
        child_b = self._pmx(parent_b.copy(), parent_a)

        return child_a, child_b

    def _pmx(self, child, parent_to_cross):
        x_indexes = np.sort(np.random.randint(0, len(child), size=2))
        
        for index in range(x_indexes[0], x_indexes[1]):
            city = parent_to_cross[index]
            swap_index = np.where(child == city)[0][0]
            child[index], child[swap_index] = child[swap_index], child[index]

        return child
            

class TruncationSelector(AbstractSelector):
    '''
    Simple truncation selection of the mew fittest 
    individuals in the population
    '''
    def __init__(self, mew):
        self._mew = mew

    def select(self, population, fitness):
        fittest_indexes = np.argpartition(fitness, fitness.size - self._mew)[-self._mew:]
        return population[fittest_indexes]

class TournamentSelector(AbstractSelector):
    '''
    Encapsulates a popular GA selection algorithm called
    Tournament Selection.  An individual is selected at random
    (with replacement) as the best from the population and competes against
    a randomly selected (with replacement) challenger.  If the individual is
    victorious they compete in the next round.  If the challenger is successful
    they become the best and are carried forward to the next round. This is repeated
    for t rounds.  Higher values of t are more selective.  
    '''
    def __init__(self, tournament_size=2):
        '''
        Constructor

        Parameters:
        ---------
        tournament_size, int, must be >=1, (default=2)
        '''
        if tournament_size < 1:
            raise ValueError('tournamant size must int be >= 1')
        
        self._tournament_size = tournament_size
        
    def select(self, population, fitness):
        '''
        Select individual from population for breeding using
        a tournament approach.  t tournaments are conducted.

        Parameters:
        ---------
        population -    numpy.array.  Matrix of chromosomes 
        fitness -       numpy.array, vector of floats representing the
                        fitness of individual chromosomes

        Returns:
        --------
        numpy.array, vector (1D array) representing the chromosome
        that won the tournament.

        '''

        tournament_participants = np.random.randint(0, population.shape[0], size=self._tournament_size)
        best = population[np.argmax(fitness[tournament_participants])]

        return best


   

class TwoCityMutator(AbstractMutator):
    '''
    Mutates an individual tour by
    randomly swapping two cities.

    Designed to work with the TSP.
    '''
    def mutate(self, tour):
        '''
        Randomly swap two cities
        
        Parameters:
        --------
        tour, np.array, tour

        '''
        #remember that index 0 and len(tour) are start/end city
        to_swap = np.random.randint(0, len(tour) - 1, 2)

        tour[to_swap[0]], tour[to_swap[1]] = \
            tour[to_swap[1]], tour[to_swap[0]]

        return tour


class TwoOptMutator(AbstractMutator):
    '''
    Mutates an individual tour by
    randomly swapping two cities.

    Designed to work with the TSP
    '''
    def mutate(self, tour):
        '''
        Randomly reverse a section of the route
        
        Parameters:
        --------
        tour, np.array, tour

        '''
        #remember that index 0 and len(tour) are start/end city
        to_swap = np.random.randint(0, len(tour) - 1, 2)

        if to_swap[1] < to_swap[0]:
            to_swap[0], to_swap[1] = to_swap[1], to_swap[0]

        return self._reverse_sublist(tour, to_swap[0], to_swap[1])


    def _reverse_sublist(self, lst, start, end):
        """
        Reverse a slice of the @lst elements between
        @start and @end
        """
        lst[start:end] = lst[start:end][::-1]
        return lst


class GeneticAlgorithmStrategy(AbstractEvolutionStrategy):
    '''
    The Genetic evolution
    Individual chromosomes in the population
    compete to cross over and breed children.  
    Children are randomly mutated.

    Each generation is of size lambda.
    '''
    def __init__(self, _lambda, selector, xoperator, mutator):
        '''
        Constructor

        Parameters:
        --------

        _lambda -   int, controls the size of each generation. (make it even)

        selector -  AbstractSelector, selects an individual chromosome for crossover

        xoperator - AbstractCrossoverOperator, encapsulates the logic
                    crossover two selected parents
        
        mutator -   AbstractMutator, encapsulates the logic of mutation for a 
                    selected individual
        '''
 
        self._lambda = _lambda
        self._selector = selector
        self._xoperator = xoperator
        self._mutator = mutator

    
    def evolve(self, population, fitness):
        '''
        Truncation selection - only mew fittest individuals survive.  
        
        Each of these breed lambda/mew children who are mutations
        of the parent.

        Parameters:
        --------
        population -- numpy array, matrix representing a generation of tours
                      size (lambda, len(individual))

        fitness --    numpy.array, vector, size lambda, representing the cost of the 
                      tours in population

        Returns:
        --------
        numpy.array - matrix of new generation, size (lambda, len(individual))
        '''

        next_gen = np.full((self._lambda, len(population[0])),
                             fill_value=-1, dtype=np.byte)

        index = 0
        for crossover in range(int(self._lambda / 2)):
            
            parent_a = self._selector.select(population, fitness)
            parent_b = self._selector.select(population, fitness)
            
            c_a, c_b = self._xoperator.crossover(parent_a, parent_b)
           
            self._mutator.mutate(c_a)
            self._mutator.mutate(c_b)

            next_gen[index], next_gen[index+1] = c_a, c_b
            
            index += 2
        return next_gen


class ElitistGeneticAlgorithmStrategy(AbstractEvolutionStrategy):
    '''
    The Genetic evolution
    Individual chromosomes in the population
    compete to cross over and breed children.  
    Children are randomly mutated.

    Each generation is of size lambda.
    '''
    def __init__(self, mew, _lambda, selector, xoperator, mutator):
        '''
        Constructor

        Parameters:
        --------

        _lambda -   int, controls the size of each generation. (make it even)

        selector -  AbstractSelector, selects an individual chromosome for crossover

        xoperator - AbstractCrossoverOperator, encapsulates the logic
                    crossover two selected parents
        
        mutator -   AbstractMutator, encapsulates the logic of mutation for a 
                    selected individual
        '''
 
        self._mew = mew
        self._lambda = _lambda
        self._selector = selector
        self._xoperator = xoperator
        self._mutator = mutator
        self._trunc_selector = TruncationSelector(mew)

    
    def evolve(self, population, fitness):
        '''
        Truncation selection - only mew fittest individuals survive.  
        
        Each of these breed lambda/mew children who are mutations
        of the parent.

        Parameters:
        --------
        population -- numpy array, matrix representing a generation of tours
                      size (lambda, len(individual))

        fitness --    numpy.array, vector, size lambda, representing the cost of the 
                      tours in population

        Returns:
        --------
        numpy.array - matrix of new generation, size (lambda, len(individual))
        '''

        next_gen = np.full((self._mew + self._lambda, len(population[0])),
                             fill_value=-1, dtype=np.byte)


        #the n fittest chromosomes in the population (breaking ties at random)
        #this is the difference from the standard GA strategy
        fittest = self._trunc_selector.select(population, fitness)
        next_gen[:len(fittest),] = fittest                     

        index = self._mew
        for crossover in range(int(self._lambda / 2)):
            
            parent_a = self._selector.select(population, fitness)
            parent_b = self._selector.select(population, fitness)
            
            c_a, c_b = self._xoperator.crossover(parent_a, parent_b)
           
            self._mutator.mutate(c_a)
            self._mutator.mutate(c_b)

            next_gen[index], next_gen[index+1] = c_a, c_b
            
            index += 2
        return next_gen


class MuLambdaEvolutionStrategy(AbstractEvolutionStrategy):
    '''
    The (mu, lambda) evolution strategy
    The fittest mew of each generation
    produces lambda/mew offspring each of which is
    mutated.

    Each generation is of size lambda.
    '''
    def __init__(self, mu, _lambda, mutator):
        '''
        Constructor

        Parameters:
        --------
        mu -       int, controls how selectiveness the algorithm.  
                    Low values of mew relative to _lambsa mean that only the best 
                    breed in each generation and the algorithm becomes 
                    more exploitative.

        _lambda -   int, controls the size of each generation.


        mutator -   AbstractMutator, encapsulates the logic of mutation for a 
                    selected individual
        '''
        self._mu = mu
        self._lambda = _lambda
        self._selector = TruncationSelector(mu)
        self._mutator = mutator

    
    def evolve(self, population, fitness):
        '''
        Truncation selection - only mew fittest individuals survive.  
        
        Each of these breed lambda/mew children who are mutations
        of the parent.

        Parameters:
        --------
        population -- numpy array, matrix representing a generation of tours
                      size (lambda, len(individual))

        fitness --    numpy.array, vector, size lambda, representing the cost of the 
                      tours in population

        Returns:
        --------
        numpy.array - matrix of new generation, size (lambda, len(individual))
        '''

        fittest = self._selector.select(population, fitness)
        population = np.full((self._lambda, fittest[0].shape[0]),
                             fill_value=-1, dtype=np.byte)

        index = 0
        for parent in fittest:
            for child_n in range(int(self._lambda/self._mu)):
                child = self._mutator.mutate(parent.copy())
                population[index] = child
                index += 1

        return population
        

class MuPlusLambdaEvolutionStrategy(AbstractEvolutionStrategy):
    '''
    The (mu+lambda) evolution strategy
    The fittest mew of each generation
    produces lambda/mew offspring each of which is
    mutated.  The mew fittest parents compete with 
    their offspring int he new generation.

    The first generation is of size lambda.
    The second generation is of size mew+lambda
    '''
    def __init__(self, mu, _lambda, mutator):
        '''
        Constructor

        Parameters:
        --------
        mu -       int, controls how selectiveness the algorithm.  
                    Low values of mew relative to _lambsa mean that only the best 
                    breed in each generation and the algorithm becomes 
                    more exploitative.

        _lambda -   int, controls the size of each generation.

        mutator -   AbstractMutator, encapsulates the logic of mutation for an indiviudal
        '''
        self._mu = mu
        self._lambda = _lambda
        self._selector = TruncationSelector(mu)
        self._mutator = mutator

    
    def evolve(self, population, fitness):
        '''
        Only mew fittest individual survice.
        Each of these breed lambda/mew children who are mutations
        of the parent.

        Parameters:
        --------
        population -- numpy array, matrix representing a generation of tours
                      size (lambda+mu, len(tour))

        fitness     -- numpy.array, vector, size lambda, representing the fitness of the 
                       individuals in the population

        Returns:
        --------
        numpy.arrays - matric a new generation of individuals, 
                       size (lambda+mu, len(individual))
        '''

        fittest = self._selector.select(population, fitness)
        
        #this is the difference from (mew, lambda)
        #could also use np.empty - quicker for larger populations...
        population = np.full((self._lambda+self._mu, fittest[0].shape[0]),
                             0, dtype=np.byte)

        population[:len(fittest),] = fittest
    
        index = self._mu
        for parent in range(len(fittest)):
            for child_n in range(int(self._lambda/self._mu)):
                child = self._mutator.mutate(fittest[parent].copy())
                population[index] = child
                index += 1
            
        return population


    
class EvolutionaryAlgorithm(object):
    '''
    Encapsulates a simple Evolutionary algorithm
    with mutation at each generation.
    '''
    def __init__(self, initialiser, objective, _lambda, 
                 strategy, maximisation=True, generations=1000):
        '''
        Parameters:
        ---------
        tour        - np.array, cities to visit
        matrix      - np.array, cost matrix travelling from city i to city j
        _lambda     - int, initial population size
        strategy    - AbstractEvolutionStrategy, evolution stratgy
        maximisation- bool, True if the objective is a maximisation and 
                      False if objective is minimisation (default=True)
        generations - int, maximum number of generations  (default=1000)
        '''
        self._initialiser = initialiser
        self._max_generations = generations
        self._objective = objective
        self._strategy = strategy
        self._lambda = _lambda
        self._best = None
        self._best_fitness = np.inf
        
        if maximisation:
            self._negate = 1
        else:
            self._negate = -1

    def _get_best(self):
        return self._best
    
    def _get_best_fitness(self):
        return self._best_fitness * self._negate

    def solve(self):

        #population = initiation_population(self._lambda, self._tour)
        population = self._initialiser.generate(self._lambda)
        fitness = None
    
        for generation in range(self._max_generations):
            fitness = self._fitness(population)
            
            max_index = np.argmax(fitness)

            if self._best is None or (fitness[max_index] > self._best_fitness):
                self._best = population[max_index]
                self._best_fitness = fitness[max_index]
            
            population = self._strategy.evolve(population, fitness)
            

    
    def _fitness(self, population):
        fitness = np.full(len(population), -1.0, dtype=float)
        for i in range(len(population)):
            
            #specific to the TSP - needs to be encapsulated...
            #fitness[i] = tour_cost(population[i], self._matrix)
            fitness[i] = self._objective.evaluate(population[i])

        return fitness * self._negate
            
    best_solution = property(_get_best)
    best_fitness = property(_get_best_fitness)



class BasicFacilityLocationMutator(AbstractMutator):
    '''
    Mutates an individual solution by
    randomly resampling each element with constant probability

    Designed to work with Facility location problems
    
    A recommended mutation rate is 1/n where n is the solution size
    '''
    
    def __init__(self, n_candidates, solution_size, mutation_rate=None,
                 verbose=False):
        '''
        Init the mutator
        
        Parameters:
        -----------
        n_candidates: int
            Number of discrete locations available
            
        solution_size: int
            Length of a solution vector (chromosome)
            
        mutation_rate: float (optional, default = None)
            If None then set to 1 / solution_size
            
        verbose: bool (optional default = False)
            Useful for education only.  Prints out each stage of the mutation.

        '''
        self.rng = np.random.default_rng()
        self.solution_size = solution_size
        self.n_candidates = n_candidates
        self.verbose = verbose
        
        if mutation_rate is None:
            self.mutation_rate = 1 / solution_size
        else:    
            self.mutation_rate = mutation_rate
    
    def mutate(self, solution):
        '''
        Randomly mutate a facility location solution
        
        Parameters:
        --------
        solution, np.ndarray, 
            A facility allocation e.g. [10, 2, 15, 20]

        '''
        #1 = mutate element
        mask = self.rng.binomial(n=1, p=self.mutation_rate, 
                                 size=len(solution))
        
        if self.verbose:
            print(f'elements mutated: {mask}')
        
        #elements keep the same (reverse of mask)
        mutant = solution[~mask.astype(bool)]
        
        if self.verbose:
            print(f'mutant_part1: {mutant}')
        
        #remaining candidate solutions to sample from...
        candidates = np.arange(self.n_candidates)
        
        if self.verbose:
            print(f'candidates1: {candidates}')
        
        mask = np.isin(candidates, solution)
        candidates = candidates[~mask]
        
        if self.verbose:
            print(f'candidates: {candidates}')
    
        
        #new locations to add to mutant
        new_locations = self.rng.choice(candidates, 
                                        size=self.solution_size - len(mutant), 
                                        replace=False)      
        if self.verbose:
            print(f'new_locations: {new_locations}')
            
        #concate the mutant 
        mutant = np.concatenate([mutant, new_locations])

        return mutant


            
class FacilityLocationPopulationGenerator(AbstractPopulationGenerator):
    '''
    Logic for generating a random finite sized population
    of facility locations.
    '''
    def __init__(self, n_candidates, n_facilities, random_seed=None):
        self.n_candidates = n_candidates
        self.n_facilities = n_facilities
        self.rng = np.random.default_rng(random_seed)

    def generate(self, population_size):
        '''
        Generate a list of @population_size solutions. Solutions
        are randomly generated and unique to maximise
        diversity of the population.

        Parameters:
        ---------
        population_size -- the size of the population

        Returns:
        ---------
        np.ndarray. 
        
            matrix size = (population_size, n_facilities). 
            Contains the initial generation of facility locations
        '''

        #fast lookup to check if solution already exists
        population = {}

        #return data as
        population_arr = np.full((population_size, self.n_facilities), -1, 
                                 dtype=np.byte)

        i = 0
        while i < population_size:
            #sample a permutation
            new_solution = self.random_solution()
            
            #check its unique to maximise diversity
            if str(new_solution) not in population:
                population[str(new_solution)] = new_solution
                i = i + 1

        #save unique permutation
        population_arr[:,] = list(population.values())

        return population_arr
    
    
    def random_solution(self):
        '''
        construct a random solution for the facility location
        problem.  Returns vector of length p
        '''
        #create array of candidate indexes
        candidates = np.arange(self.n_candidates, dtype=np.byte)

        #sample without replacement and return array
        return self.rng.choice(candidates, size=self.n_facilities, 
                               replace=False)



class WeightedAverageObjective:
    '''
    Encapsulates logic for calculation of
    weighted average in a simple facility location problem

    Thanks:
    -------
    Thanks goes to Sammi and Ben - two students in the 2020/21 
    MSc HDS cohort for finding and fixing a bug in the evaluate() method.
    '''

    def __init__(self, demand, travel_matrix):
        '''
        Constructor. store the demand and travel times used
        to calculate the weighted average.

        Parameters:
        ---------
        demand: pd.DataFrame
            demand for by geo sector.  Assumes includes cols
            'sector' and 'n_patients'

        travel_matrix: pd.DataFrame
            Travel cost from geo sector (row) to facility (col)
        '''
        self.demand = demand
        self.travel_matrix = travel_matrix

    def evaluate(self, solution):
        '''
        Calculate the weighted average travel time for a solution.
        Assumes patients travel to closest facility.

        Parameters:
        ----------
        solution: np.array
            Representation of a solution
            e.g. [0, 1, 2, 3, 4] will select the first 5 facilities

        Returns:
        --------
            float
        '''

        # only select clinics encoded with 1 in the solution (cast to bool)
        mask = self.travel_matrix.columns[solution]
        active_facilities = self.travel_matrix[mask]

        # turn off chained assignment: avoid displaying false +ive
        # SettingWithCopy warning.
        pd.set_option('mode.chained_assignment', None)

        # assume travel to closest facility
        active_facilities['min_cost'] = \
            active_facilities.min(axis=1).to_numpy()

        # turn warnings back on for chained assignment.
        pd.set_option('mode.chained_assignment', 'warn')

        # merge demand and travel times into a single DataFrame
        problem = self.demand.merge(active_facilities, on='sector',
                                    how='inner')

        # return weighted average
        return np.average(problem['min_cost'],
                          weights=problem['n_patients'])



class FacilityLocationSinglePointCrossOver():
    '''
    Single point cross over for a facility location problem
    with non-binary representation.
    '''
    def __init__(self):
        self.rng = np.random.default_rng()
    
    def crossover(self, parent_a, parent_b):
        
        #generate exchange vectors
        ex_vector_a = parent_a[~np.isin(parent_a, parent_b)]
        ex_vector_b = parent_b[~np.isin(parent_b, parent_a)]
        
        #cross over points
        #copy parents if equal and no exchange possible.
        if len(ex_vector_a) > 0:
            x_point = self.rng.integers(len(ex_vector_a))

            #child a
            child_a = parent_a[np.isin(parent_a, parent_b)]
            child_a = np.concatenate([child_a, 
                                      ex_vector_a[:x_point],
                                      ex_vector_b[x_point:]])           
            #child b
            child_b = parent_b[np.isin(parent_b, parent_a)]
            child_b = np.concatenate([child_b, 
                                      ex_vector_b[:x_point],
                                      ex_vector_a[x_point:]])
        else:
            child_a = parent_a
            child_b = parent_b
               
        return child_a, child_b



    





# -*- coding: utf-8 -*-
"""
Provides euclidean functions for working with TSP data.

One key function converts a numpy array of 2D coordinates into a matrix of euclidean distances
between them.
"""

import numpy as np
import matplotlib.pyplot as plt
from decimal import Decimal, ROUND_HALF_UP

def gen_matrix(cities, as_integer=False):
    """
    Creates a numpy array of euclidian distances between 2 sets of
    cities
    
    Parameters:
    ----------
    points: numpy.array
        coordinate pairs
    
    as_integers: bool, optional (default=False)
        If true then round to nearest int
        Behaviour: 1.5 -> 2
                   1.2 -> 1
                   1.8 -> 2
                   
    Returns:
    -------
        np.ndarray
            Matrix of city to city costs
    
    """
    size = len(cities)
    matrix = np.zeros(shape=(size, size))
    
    row = 0
    col = 0
    
    for city1 in cities:
        col = 0
        for city2 in cities:
            distance = euclidean_distance(city1, city2)
            if as_integer:
                distance = int(Decimal(distance).quantize(0, ROUND_HALF_UP))
            matrix[row, col] = distance
            col+=1
        row +=1
        
    return matrix


def euclidean_distance(city1, city2):
    """
    Calculate euc distance between 2 cities
    5.5 ms to execute
    """
    return np.linalg.norm((city1-city2))



def euclidean_distance2(city1, city2):
    """
    An alterantive way to calculate euc distance 
    between two cities
    
    6-7 ms to execute
    """
    v = np.square(city1 - city2)
    return np.sqrt(np.sum(v))


def plot_tour(tour, cities, figsize=(6,4)):
    '''
    Plots a tour.  Each city visited is
    labelled in order.  Red point is the initial city.
    
    Params:
    ------
    tour: np.ndarray
        ordered vector representing tour e.g. [1, 4, 2, 3]
        
    cities: np.ndarray
        matrix representing city coordinates
        
    figsize: tuple, optional (default = (6,3))
        tuple of ints for figure size
        
    Returns
    -------
        tuple of matplotlib figure, and axis
    '''

    tour_length = len(tour)
    fig, ax = plt.subplots(1, 1, figsize=figsize)

    #plot points
    ax.plot(cities[:tour_length][:, 0],
            cities[:tour_length][:, 1],'bo')

    #plot lines
    for j in range(len(tour)-1):
        city_1 = tour[j]
        city_2 = tour[j+1]

        #lookup coordinates
        coords_1 = cities[city_1]
        coords_2 = cities[city_2]
        coords = np.vstack([coords_1, coords_2])

        #plot lines
        ax.plot(coords[:,0], coords[:,1], 'g-')
        
        #show order in tour
        ax.text(coords_1[0] + 0.8, coords_1[1] + 0.8, str(j))

    #add in loop back colour code in red...
    city_1 = tour[-1]
    city_2 = tour[0]

    coords_1 = cities[city_1]
    coords_2 = cities[city_2]
    coords = np.vstack([coords_1, coords_2])
    ax.text(coords_1[0] + 0.8, coords_1[1] + 0.8, str(tour_length-1))
    ax.plot(coords[:,0], coords[:,1], 'r--')
        
    return fig, ax


# -*- coding: utf-8 -*-
"""
Functions and classes to enable bruteforce solution of the TSP.
Note Bruteforce is inefficient after tours exceed 5 cities
"""


import itertools as ite
import numpy as np
from sympy.utilities.iterables import multiset_permutations

from metapy.tsp.objective import tour_cost
from metapy.tsp.init_solutions import random_tour
from metapy.tsp.tsp_utility import append_base, trim_base
    
class BruteForceSolver(object):
    """
    Enumerates all permutations of a given list of cities and calculates
    waits. Note for n cities there are n! permutations!  BruteForceSolver is
    too slow for tours beyond 10 cities.
    """
    def __init__(self, init_solution, objective, maximisation=False):
        """
        Constructor Method
        
        Params:
        -------
        init_solution, np.ndarray
            initial tour
        objective, Object
            Class that implements .evaluate() interface
        """
        self._objective = objective
        self._init_solution = init_solution

        if maximisation:
            self._negate = 1.0
        elif not maximisation:
            self._negate = -1.0
        else:
            raise ValueError('parameter maximisation must be bool True|False')

        #list of best solutions in case of multiple optima
        self.best_solutions = [init_solution]
        self._best_cost = self._objective.evaluate(init_solution) * self._negate

    def _get_best_cost(self):
        return self._best_cost * self._negate
       
    def solve(self):
        """
        Enumerate all costs to find the minimum.
        Store solution(s)
        """
    
        #pick a random start city and then permute the rest and rejoin.
        origin = np.array([self._init_solution[0]])
        for current_tour in ite.permutations(self._init_solution[1:]):
            current_tour = np.concatenate([origin, current_tour])
            cost = self._objective.evaluate(current_tour) * self._negate
                        
            if self._best_cost == cost:
                self._best_cost = cost
                self.best_solutions.append(current_tour)
                
            elif cost > self._best_cost:
                self._best_cost = cost
                self.best_solutions = [current_tour]

    def _all_permutations(self, tour):
        """
        Returns a list of lists containing all permutations of a
        tour.  The base_city is appended to each_list
        """
        return [np.array(x) for x in ite.permutations(tour)]

    best_cost = property(_get_best_cost)
        
                
                
            
class RandomSearch(object):
    """
    A simple global optimisation algorithm - encapsulates Random Search.  
    The algorithm is completely explorative and randomly
    samples a tour and compares if it is better than the current
    best.
    """
    def __init__(self, init_solution, objective, max_iter=1000, maximisation=False):
        """
        Constructor Method

        Parameters:
        ---------
        init_solution -- initial tour
        matrix -- matrix of travel costs
        """
        self._objective = objective
        self._init_solution = init_solution
        self._max_iter = max_iter

        if maximisation:
            self._negate = 1.0
        elif not maximisation:
            self._negate = -1.0
        else:
            raise ValueError('parameter maximisation must be bool True|False')

        #list of best solutions in case of multiple optima
        self.best_solutions = [init_solution]
        self._best_cost = self._objective.evaluate(init_solution) * self._negate

    def _get_best_cost(self):
        return self._best_cost * self._negate
        
    def solve(self):
        '''
        Random search.
        
        Loop until all iterations are complete.  
        Sample a random tour on each iterations and compare to best.
        
        '''
        
        for iteration in range(self._max_iter):
            sample_tour = np.random.permutation(self._init_solution)
            sample_cost = self._objective.evaluate(sample_tour) * self._negate

            if self.best_cost == sample_cost:
                self.best_solutions.append(sample_tour)

            elif sample_cost > self._best_cost:
                self.best_solutions = [sample_tour]
                self._best_cost = sample_cost
        
    best_cost = property(_get_best_cost)
